import copy
import logging
from typing import Union

import numpy as np
from pandas import DataFrame, Series

from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, QUANTILE, SOFTCLASS

from autogluon.core.utils import get_pred_from_proba, get_pred_from_proba_df

logger = logging.getLogger(__name__)


# LabelCleaner cleans labels prior to entering feature generation
class LabelCleaner:
    num_classes = None
    inv_map = None
    ordered_class_labels = None
    ordered_class_labels_transformed = None
    problem_type_transform = None

    @staticmethod
    def construct(problem_type: str, y: Union[Series, np.ndarray, list, DataFrame], y_uncleaned: Union[Series, np.ndarray, list, DataFrame] = None, positive_class=None):
        if problem_type == SOFTCLASS:
            return LabelCleanerSoftclass(y)
        y = LabelCleaner._convert_to_valid_series(y)
        if y_uncleaned is not None:
            y_uncleaned = LabelCleaner._convert_to_valid_series(y_uncleaned)

        if problem_type == BINARY:
            return LabelCleanerBinary(y, positive_class=positive_class)
        elif problem_type == MULTICLASS:
            if y_uncleaned is None:
                y_uncleaned = copy.deepcopy(y)
            if len(y.unique()) == 2:
                return LabelCleanerMulticlassToBinary(y, y_uncleaned)
            else:
                return LabelCleanerMulticlass(y, y_uncleaned)
        elif problem_type in [REGRESSION, QUANTILE]:
            return LabelCleanerDummy(problem_type=problem_type)
        else:
            raise NotImplementedError

    def transform(self, y: Union[Series, np.ndarray, list]) -> Series:
        y = self._convert_to_valid_series(y)
        return self._transform(y)

    def inverse_transform(self, y: Union[Series, np.ndarray, list]) -> Series:
        y = self._convert_to_valid_series(y)
        return self._inverse_transform(y)

    def _transform(self, y: Series) -> Series:
        raise NotImplementedError

    def _inverse_transform(self, y: Series) -> Series:
        raise NotImplementedError

    def transform_proba(self, y: Union[DataFrame, Series, np.ndarray], as_pandas=False):
        return y

    def inverse_transform_proba(self, y, as_pandas=False, as_pred=False):
        return y

    @staticmethod
    def _convert_to_valid_series(y: Union[Series, np.ndarray, list]) -> Series:
        if isinstance(y, np.ndarray) or isinstance(y, list):
            y = Series(y)
        elif isinstance(y, Series) and y.dtype.name == 'category':
            y = y.astype('object')
        return y


class LabelCleanerMulticlass(LabelCleaner):
    def __init__(self, y: Series, y_uncleaned: Series):
        self.problem_type_transform = MULTICLASS
        y = self._convert_to_valid_series(y)
        y_uncleaned = self._convert_to_valid_series(y_uncleaned)
        self.cat_mappings_dependent_var: dict = self._generate_categorical_mapping(y)
        self.inv_map: dict = {v: k for k, v in self.cat_mappings_dependent_var.items()}

        self.cat_mappings_dependent_var_uncleaned: dict = self._generate_categorical_mapping(y_uncleaned)
        self.inv_map_uncleaned: dict = {v: k for k, v in self.cat_mappings_dependent_var_uncleaned.items()}

        self.num_classes = len(self.cat_mappings_dependent_var.keys())
        self.ordered_class_labels = list(y_uncleaned.astype('category').cat.categories)
        self.valid_ordered_class_labels = list(y.astype('category').cat.categories)
        self.ordered_class_labels_transformed = list(range(len(self.valid_ordered_class_labels)))
        self.invalid_class_count = len(self.ordered_class_labels) - len(self.valid_ordered_class_labels)
        self.labels_to_zero_fill = [1 if label not in self.valid_ordered_class_labels else 0 for label in self.ordered_class_labels]
        self.label_index_to_keep = [i for i, label in enumerate(self.labels_to_zero_fill) if label == 0]
        self.label_index_to_remove = [i for i, label in enumerate(self.labels_to_zero_fill) if label == 1]

    def _transform(self, y: Series) -> Series:
        y = y.map(self.inv_map)
        return y

    def _inverse_transform(self, y: Series) -> Series:
        y = y.map(self.cat_mappings_dependent_var)
        return y

    # TODO: Unused? There are not many reasonable situations that seem to require this method.
    def transform_proba(self, y: Union[DataFrame, np.ndarray], as_pandas=False) -> Union[DataFrame, np.ndarray]:
        if self.invalid_class_count > 0:
            # this assumes y has only 0's for any columns it is about to remove, if it does not, weird things may start to happen since rows will not sum to 1
            if isinstance(y, DataFrame):
                cols_to_drop = [self.cat_mappings_dependent_var_uncleaned[col] for col in self.label_index_to_remove]
                y = y.drop(columns=cols_to_drop)
            else:
                y = np.delete(y, self.label_index_to_remove, axis=1)
        if as_pandas:
            if isinstance(y, DataFrame):
                return y.rename(columns=self.inv_map)
            else:
                return DataFrame(data=y, columns=self.ordered_class_labels_transformed, dtype=np.float32)
        elif isinstance(y, DataFrame):
            return y.to_numpy()
        else:
            return y

    def inverse_transform_proba(self, y, as_pandas=False, as_pred=False):
        y_index = None
        if isinstance(y, DataFrame):
            y_index = y.index
            y = y.to_numpy()
        if self.invalid_class_count > 0:
            y_transformed = np.zeros([len(y), len(self.ordered_class_labels)], dtype=np.float32)
            y_transformed[:, self.label_index_to_keep] = y
        else:
            y_transformed = y
        if as_pred:
            y_transformed = get_pred_from_proba(y_pred_proba=y_transformed, problem_type=MULTICLASS)
            y_transformed = self._convert_to_valid_series(y_transformed)
            y_transformed = y_transformed.map(self.cat_mappings_dependent_var_uncleaned)
            if y_index is not None:
                y_transformed.index = y_index
        if as_pandas and not as_pred:
            y_transformed = DataFrame(data=y_transformed, index=y_index, columns=self.ordered_class_labels, dtype=np.float32)
        return y_transformed

    @staticmethod
    def _generate_categorical_mapping(y: Series) -> dict:
        categories = y.astype('category')
        cat_mappings_dependent_var = dict(enumerate(categories.cat.categories))
        return cat_mappings_dependent_var


# TODO: Expand print statement to multiclass as well
class LabelCleanerBinary(LabelCleaner):
    def __init__(self, y: Series, positive_class=None):
        self.problem_type_transform = BINARY
        y = self._convert_to_valid_series(y)
        self.num_classes = 2
        self.unique_values = list(y.unique())
        if len(self.unique_values) != 2:
            raise AssertionError('y does not contain exactly 2 unique values:', self.unique_values)
        try:
            self.unique_values.sort()
        except TypeError:
            # Contains both str and int
            self.unique_values = list(y.unique())
        # TODO: Clean this code, for loop
        pos_class_warning = None
        if positive_class is not None:
            if positive_class not in self.unique_values:
                raise ValueError(f'positive_class is not a valid class: {self.unique_values} (positive_class={positive_class})')
            negative_class = [c for c in self.unique_values if c != positive_class][0]
            self.inv_map: dict = {negative_class: 0, positive_class: 1}
        elif ((str(False) in [str(val) for val in self.unique_values]) and
              (str(True) in [str(val) for val in self.unique_values])):
            false_val = [val for val in self.unique_values if str(val) == str(False)][0]  # may be str or bool
            true_val = [val for val in self.unique_values if str(val) == str(True)][0]  # may be str or bool
            self.inv_map: dict = {false_val: 0, true_val: 1}
        elif (0 in self.unique_values) and (1 in self.unique_values):
            self.inv_map: dict = {0: 0, 1: 1}
        else:
            self.inv_map: dict = {self.unique_values[0]: 0, self.unique_values[1]: 1}
            pos_class_warning = f'\tNote: For your binary classification, AutoGluon arbitrarily selected which label-value ' \
                                f'represents positive ({self.unique_values[1]}) vs negative ({self.unique_values[0]}) class.\n'\
                                '\tTo explicitly set the positive_class, either rename classes to 1 and 0, or specify positive_class in Predictor init.'
        poslabel = [lbl for lbl in self.inv_map.keys() if self.inv_map[lbl] == 1][0]
        neglabel = [lbl for lbl in self.inv_map.keys() if self.inv_map[lbl] == 0][0]
        logger.log(20, 'Selected class <--> label mapping:  class 1 = %s, class 0 = %s' % (poslabel, neglabel))
        if pos_class_warning is not None:
            logger.log(20, pos_class_warning)
        self.cat_mappings_dependent_var: dict = {v: k for k, v in self.inv_map.items()}
        self.ordered_class_labels_transformed = [0, 1]
        self.ordered_class_labels = [self.cat_mappings_dependent_var[label_transformed] for label_transformed in self.ordered_class_labels_transformed]

    def inverse_transform_proba(self, y, as_pandas=False, as_pred=False):
        if isinstance(y, DataFrame):
            y = copy.deepcopy(y)
            y.columns = copy.deepcopy(self.ordered_class_labels)
            if as_pred:
                y = get_pred_from_proba_df(y, problem_type=self.problem_type_transform)
            if not as_pandas:
                y = y.to_numpy()
        elif as_pred:
            y_index = None
            if isinstance(y, Series):
                y_index = y.index
                y = y.to_numpy()
            y = get_pred_from_proba(y_pred_proba=y, problem_type=self.problem_type_transform)
            y = self._convert_to_valid_series(y)
            y = y.map(self.cat_mappings_dependent_var)
            y = y.to_numpy()
            if as_pandas:
                y = Series(data=y, index=y_index)
        return y

    def _transform(self, y: Series) -> Series:
        y = y.map(self.inv_map)
        return y

    def _inverse_transform(self, y: Series) -> Series:
        return y.map(self.cat_mappings_dependent_var)

    def transform_proba(self, y: Union[DataFrame, Series, np.ndarray], as_pandas=False):
        if not as_pandas and isinstance(y, (Series, DataFrame)):
            y = y.to_numpy()
        elif isinstance(y, DataFrame):
            y = copy.deepcopy(y)
            y.columns = copy.deepcopy(self.ordered_class_labels_transformed)
        return y


class LabelCleanerMulticlassToBinary(LabelCleanerMulticlass):
    def __init__(self, y: Series, y_uncleaned: Series):
        super().__init__(y=y, y_uncleaned=y_uncleaned)
        self.label_cleaner_binary = LabelCleanerBinary(y=y.map(self.inv_map))
        self.problem_type_transform = self.label_cleaner_binary.problem_type_transform

    def _transform(self, y: Series) -> Series:
        y = super()._transform(y)
        y = self.label_cleaner_binary.transform(y)
        return y

    def inverse_transform_proba(self, y, as_pandas=False, as_pred=False):
        if not isinstance(y, DataFrame):
            y = self.convert_binary_proba_to_multiclass_proba(y=y, as_pandas=as_pandas)
        return super().inverse_transform_proba(y, as_pandas=as_pandas, as_pred=as_pred)

    @staticmethod
    def convert_binary_proba_to_multiclass_proba(y, as_pandas=False):
        y_index = None
        if as_pandas and isinstance(y, Series):
            y_index = y.index
        y_transformed = np.zeros([len(y), 2])
        y_transformed[:, 0] = 1 - y
        y_transformed[:, 1] = y
        if as_pandas:
            y_transformed = DataFrame(data=y_transformed, index=y_index)
        return y_transformed


# TODO: Expand functionality if necessary
class LabelCleanerSoftclass(LabelCleaner):
    def __init__(self, y: DataFrame):
        self.problem_type_transform = SOFTCLASS
        self.num_classes = y.shape[1]

    def _transform(self, y: DataFrame) -> DataFrame:
        return y

    def _inverse_transform(self, y: DataFrame) -> DataFrame:
        return y

    @staticmethod
    def _convert_to_valid_series(y: DataFrame) -> DataFrame:
        return y


class LabelCleanerDummy(LabelCleaner):
    def __init__(self, problem_type=REGRESSION):
        self.problem_type_transform = problem_type

    def _transform(self, y: Union[Series, DataFrame]) -> Union[Series, DataFrame]:
        return y

    def _inverse_transform(self, y: Union[Series, DataFrame]) -> Union[Series, DataFrame]:
        return y



from abc import ABC, abstractmethod
from typing import Set

from ..datatypes.common import Candidate
from ..datatypes.hp_ranges import HyperparameterRanges


class DuplicateDetector(ABC):
    @abstractmethod
    def contains(self, existing_candidates: Set[Candidate], new_candidate: Candidate) -> bool:
        pass


class DuplicateDetectorNoDetection(DuplicateDetector):
    def contains(self, existing_candidates: Set[Candidate], new_candidate: Candidate) -> bool:
        return False  # no duplicate detection at all


class DuplicateDetectorIdentical(DuplicateDetector):
    def contains(self, existing_candidates: Set[Candidate], new_candidate: Candidate) -> bool:
        return new_candidate in existing_candidates


DUPLICATE_DETECTION_EPSILON = 1e-8


class DuplicateDetectorEpsilon(DuplicateDetector):
    def __init__(self, hp_ranges: HyperparameterRanges):
        self.hp_ranges = hp_ranges

    def contains(self, existing_candidates: Set[Candidate], new_candidate: Candidate) -> bool:
        return any(self._almost_equal(c, new_candidate) for c in existing_candidates)

    def _almost_equal(self, candidate1, candidate2):
        assert len(candidate1) == len(candidate2), (candidate1, candidate2)
        np_cand1 = self.hp_ranges.to_ndarray(candidate1)
        np_cand2 = self.hp_ranges.to_ndarray(candidate2)
        assert np_cand1.shape == np_cand2.shape, (np_cand1, np_cand2)
        return all(abs(hp1 - hp2) < DUPLICATE_DETECTION_EPSILON for hp1, hp2 in zip(np_cand1, np_cand2))



import gzip
import bz2
import lzma


compression_fn_map = {
    None: {
        'open': open,
        'extension': '',
    },
    'gzip': {
        'open': gzip.open,
        'extension': 'gz',
    },
    'bz2': {
        'open': bz2.open,
        'extension': 'bz2',
    },
    'lzma': {
        'open': lzma.open,
        'extension': 'lzma',
    },
}


def get_validated_path(filename, compression_fn=None):
    if compression_fn is not None:
        validated_path = f"{filename}.{compression_fn_map[compression_fn]['extension']}"
    else:
        validated_path = filename
    return validated_path


def get_compression_map():
    return compression_fn_map



import autograd.numpy as anp

from .constants import INITIAL_NOISE_VARIANCE, NOISE_VARIANCE_LOWER_BOUND, \
    NOISE_VARIANCE_UPPER_BOUND, DEFAULT_ENCODING
from .distribution import Gamma
from .gluon import Block
from .gluon_blocks_helpers import encode_unwrap_parameter
from .kernel import KernelFunction
from .mean import ScalarMeanFunction, MeanFunction
from .posterior_state import GaussProcPosteriorState
from .utils import create_encoding, register_parameter


class MarginalLikelihood(Block):
    """
    Marginal likelihood of Gaussian process with Gaussian likelihood

    :param kernel: Kernel function (for instance, a Matern52---note we cannot
        provide Matern52() as default argument since we need to provide with
        the dimension of points in X)
    :param mean: Mean function which depends on the input X only (by default,
        a scalar fitted while optimizing the likelihood)
    :param initial_noise_variance: A scalar to initialize the value of the
        residual noise variance
    """
    def __init__(
            self, kernel: KernelFunction, mean: MeanFunction = None,
            initial_noise_variance=None, encoding_type=None, **kwargs):
        super(MarginalLikelihood, self).__init__(**kwargs)
        if mean is None:
            mean = ScalarMeanFunction()
        if initial_noise_variance is None:
            initial_noise_variance = INITIAL_NOISE_VARIANCE
        if encoding_type is None:
            encoding_type = DEFAULT_ENCODING
        self.encoding = create_encoding(
             encoding_type, initial_noise_variance, NOISE_VARIANCE_LOWER_BOUND,
             NOISE_VARIANCE_UPPER_BOUND, 1, Gamma(mean=0.1, alpha=0.1))
        self.mean = mean
        self.kernel = kernel
        with self.name_scope():
            self.noise_variance_internal = register_parameter(
                self.params, 'noise_variance', self.encoding)

    def _noise_variance(self):
        return encode_unwrap_parameter(
            self.noise_variance_internal, self.encoding)

    def forward(self, X, y):
        """
        Actual computation of the marginal likelihood
        See http://www.gaussianprocess.org/gpml/chapters/RW.pdf, equation (2.30)

        :param X: input data of size (n, d)
        :param y: targets corresponding to X, of size (n, 1)
        """
        state = GaussProcPosteriorState(
            X, y, self.mean, self.kernel, self._noise_variance())
        return state.neg_log_likelihood()

    def param_encoding_pairs(self):
        """
        Return a list of tuples with the Gluon parameters of the likelihood and their respective encodings
        """
        own_param_encoding_pairs = [
            (self.noise_variance_internal, self.encoding)]
        return own_param_encoding_pairs + self.mean.param_encoding_pairs() + \
               self.kernel.param_encoding_pairs()

    def box_constraints_internal(self):
        """
        Collect the box constraints for all the underlying parameters
        """
        all_box_constraints = {}
        for param, encoding in self.param_encoding_pairs():
            assert encoding is not None,\
                "encoding of param {} should not be None".format(param.name)
            all_box_constraints.update(encoding.box_constraints_internal(param))
        return all_box_constraints

    def get_noise_variance(self, as_ndarray=False):
        noise_variance = self._noise_variance()
        return noise_variance if as_ndarray else anp.reshape(noise_variance, (1,))[0]

    def set_noise_variance(self, val):
        self.encoding.set(self.noise_variance_internal, val)
        
    def get_params(self):
        result = {'noise_variance': self.get_noise_variance()}
        for pref, func in [('kernel_', self.kernel), ('mean_', self.mean)]:
            result.update({
                (pref + k): v for k, v in func.get_params().items()})
        return result
    
    def set_params(self, param_dict):
        for pref, func in [('kernel_', self.kernel), ('mean_', self.mean)]:
            len_pref = len(pref)
            stripped_dict = {
                k[len_pref:]: v for k, v in param_dict.items()
                if k.startswith(pref)}
            func.set_params(stripped_dict)
        self.set_noise_variance(param_dict['noise_variance'])



import autograd.numpy as anp

from .constants import DEFAULT_ENCODING, INITIAL_WARPING, WARPING_LOWER_BOUND, \
    WARPING_UPPER_BOUND, NUMERICAL_JITTER
from .distribution import LogNormal
from .kernel import KernelFunction
from .gluon import Block
from .gluon_blocks_helpers import encode_unwrap_parameter
from .utils import create_encoding, register_parameter


class OneDimensionalWarping(Block):
    """
    Block that is responsible for the warping of a single, column
    feature x. Typically, the full data X = [x1, x2,..., xd] is in (n, d) and
    each xi is a column feature in (n, 1).

    Consider column feature x and assume that the entries of x are contained in
    the range input_range. Each entry of x is transformed by
        warping(u) = 1. - (1. - R(u)^a)^b,
    with a,b two non negative parameters learned by empirical Bayes, and R(.)
    is a linear transformation that, based on input_range, rescales the entry
    of x into [eps, 1-eps] for some small eps > 0.

    :param input_range: tuple that contains the lower and upper bounds of the
        entries of x.
    """
    def __init__(self, input_range, encoding_type=DEFAULT_ENCODING, **kwargs):
        super().__init__(**kwargs)
        self.input_range = input_range
        self.encoding = create_encoding(
            encoding_type, INITIAL_WARPING, WARPING_LOWER_BOUND,
            WARPING_UPPER_BOUND, 2, LogNormal(0.0, 0.75))
        with self.name_scope():
            self.warping_internal = register_parameter(
                self.params, 'warping', self.encoding, shape=(2,))

    def _rescale(self, x):
        """
        We linearly rescale the entries of x into [NUMERICAL_JITTER, 1-NUMERICAL_JITTER]
        In this way, we avoid the differentiability problems at 0
        :param x: np.ndarray to be rescaled
        """
        lower, upper = self.input_range
        P = (1. - 2 * NUMERICAL_JITTER)/(upper - lower)
        Q = (NUMERICAL_JITTER * (upper + lower) - lower)/(upper - lower)

        return P * x + Q

    def _warping(self):
        return encode_unwrap_parameter(self.warping_internal, self.encoding)

    def forward(self, x):
        """
        Actual computation of the warping transformation (see details above)

        :param x: input data of size (n,1)
        """
        warping = anp.reshape(self._warping(), (-1,))
        warping_a = warping[0]
        warping_b = warping[1]

        return 1. - anp.power(1. - anp.power(
            self._rescale(x), warping_a), warping_b)

    def param_encoding_pairs(self):
        """
        Return a list of tuples with the Gluon parameters of the 1-D warping
        and their respective encodings
        """
        return [(self.warping_internal, self.encoding)]
    
    def get_params(self):
        warping = anp.reshape(self._warping(), (-1,))
        return {
            'warping_a': warping[0],
            'warping_b': warping[1]}

    def set_params(self, param_dict):
        warping = [param_dict['warping_a'], param_dict['warping_b']]
        self.encoding.set(self.warping_internal, warping)


class Warping(Block):
    """
    Block that computes warping over all the columns of some input data X.
    If X is of size (n,dimension), where dimension has to be specified, a 1-D warping
    transformation is applied to each column X[:,j] with j a key in index_to_range.
    More precisely, index_to_range is a dictionary of the form
        {
            j : (lower_bound_column_j, upper_bound_column_j),
            k : (lower_bound_column_k, upper_bound_column_k),
            ....
        }
    that maps column indexes to their corresponding ranges.
    """
    def __init__(self, dimension, index_to_range, encoding_type=DEFAULT_ENCODING,
                 **kwargs):
        super().__init__(**kwargs)

        assert isinstance(index_to_range, dict)
        assert all(isinstance(r, tuple) for r in index_to_range.values())
        assert all(r[0] < r[1] for r in index_to_range.values()) # for dictionary

        self.transformations = []
        self._params_encoding_pairs = []
        self.dimension = dimension
        self.index_to_range = index_to_range

        some_are_warped = False
        for col_index in range(dimension):
            if col_index in index_to_range:
                transformation = OneDimensionalWarping(
                    index_to_range[col_index], encoding_type=encoding_type)
                # To make sure that OneDimensionalWarping will get initialized
                # and managed by Warping, we register it as a child.
                self.register_child(transformation, name=transformation.name)
                self._params_encoding_pairs += transformation.param_encoding_pairs()
                some_are_warped = True
            else:
                # if a column is not warped, we do not apply any transformation
                transformation = lambda x: x
            self.transformations.append(transformation)
        assert some_are_warped,\
            "At least one of the dimensions must be warped"

    def forward(self, X):
        """
        Actual computation of warping applied to each column of X

        :param X: input data of size (n,dimension)
        """
        warped_X = []
        for col_index, transformation in enumerate(self.transformations):
            x = X[:, col_index:(col_index+1)]
            warped_X.append(transformation(x))

        return anp.concatenate(warped_X, axis=1)

    def param_encoding_pairs(self):
        """
        Return a list of tuples with the Gluon parameters of the warping and
        their respective encodings
        """
        return self._params_encoding_pairs

    def get_params(self):
        """
        Keys are warping_a, warping_b if there is one dimension, and
        warping_a<k>, warping_b<k> otherwise.
        """
        if len(self.transformations) == 1:
            result = self.transformations[0].get_params()
        else:
            result = dict()
            for i, warping in enumerate(self.transformations):
                if isinstance(warping, OneDimensionalWarping):
                    istr = str(i)
                    for k, v in warping.get_params().items():
                        result[k + istr] = v
        return result

    def set_params(self, param_dict):
        if len(self.transformations) == 1:
            self.transformations[0].set_params(param_dict)
        else:
            transf_keys = None
            for i, warping in enumerate(self.transformations):
                if isinstance(warping, OneDimensionalWarping):
                    if transf_keys is None:
                        transf_keys = warping.get_params().keys()
                    istr = str(i)
                    stripped_dict = dict()
                    for k in transf_keys:
                        stripped_dict[k] = param_dict[k + istr]
                    warping.set_params(stripped_dict)


class WarpedKernel(KernelFunction):
    """
    Block that composes warping with an arbitrary kernel
    """
    def __init__(self, kernel: KernelFunction, warping: Warping, **kwargs):
        super().__init__(kernel.dimension, **kwargs)
        self.kernel = kernel
        self.warping = warping

    def forward(self, X1, X2):
        """
        Actual computation of the composition of warping with an arbitrary
        kernel K. If we have input data X1 and X2, of respective dimensions
        (n1, d) and (n2, d), we compute the matrix

            K(warping(X1), warping(X2)) of size (n1,n2)
            whose (i,j) entry is given by K(warping(X1[i,:]), warping(X2[j,:]))

        :param X1: input data of size (n1, d)
        :param X2: input data of size (n2, d)
        """
        warped_X1 = self.warping(X1)
        if X1 is X2:
            warped_X2 = warped_X1
        else:
            warped_X2 = self.warping(X2)
        return self.kernel(warped_X1, warped_X2)

    def diagonal(self, X):
        # If kernel.diagonal does not depend on content of X (but just its
        # size), can pass X instead of self.warping(X)
        warped_X = self.warping(X) if self.kernel.diagonal_depends_on_X()             else X
        return self.kernel.diagonal(warped_X)

    def diagonal_depends_on_X(self):
        return self.kernel.diagonal_depends_on_X()

    def param_encoding_pairs(self):
        return self.kernel.param_encoding_pairs() + \
               self.warping.param_encoding_pairs()

    def get_params(self):
        # We use the union of get_params for kernel and warping, without
        # prefixes.
        result = self.kernel.get_params()
        result.update(self.warping.get_params())
        return result

    def set_params(self, param_dict):
        self.kernel.set_params(param_dict)
        self.warping.set_params(param_dict)
